// Jason Rohrer
// Client.java

/**
*
*	konspire client 
*
*	Created 5-11-2000
*	Mods:
*		Jason Rohrer	6-16-2000	Client now randomizes order of server list vector
*									before trying to connect. 
*		Jason Rohrer	6-25-2000	Fixed bug that caused client to start rejecting connections
*									number of connected clients was being incremented instead
*									of decremented when a client disconnnected.
*		Jason Rohrer	7-4-2000	Changed so that client looks for a "good" local
*									address in the server list and replaces
*									its local address with the new address.
*									Client is responsible for updating its FileHost
*									message so that all file descriptors have the correct
*									address.
*		Jason Rohrer	7-23-2000	Changed so that all spawned threads are set to 
*									minimum priority to make sure that the GUI
*									is always completely responsive.  Note that this
*									priority setting doesn't usually increase GUI responsiveness
*									noticibly, since most threads are handling network connections
*									and are therefore blocked often enough so that the GUI thread
*									has ample time to run.  However, on faster network connections,
*									e.g., gigabit ethernet, the network threads will not
*									be blocked as often and could cause GUI responsiveness
*									to be sporadic.  We're assuming that as long as the GUI
*									isn't being manipulated or updated, MIN_PRIORITY thread
*									performance will not be affected.  This should be verified.
*		Jason Rohrer	7-24-2000	Fixed 1.1 compatability bug in menu handle.  Was using getParentFile(),
*									needed to use getParent().
*/

package konspire.client;

import konspire.common.*;

import konspire.common.FileDescriptor;

import java.net.*;
import java.io.*;
import java.awt.Event;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.WindowEvent;
import java.awt.FlowLayout;
import java.awt.Dimension;
import javax.swing.*;


import java.util.Enumeration;
import java.util.NoSuchElementException;

import java.util.Random;


/**
 * Main <code>JFrame</code> for konspire client.  Contains main method.
 *
 * @author Jason Rohrer
 */
public class Client extends JFrame implements DownloadController {
	
	static int defaultPort = 6433;
	static int defaultMaxClients = 10;
	
	static Random randGen = new Random();
	
	int port;

	int maxClients;
	int maxConnections;
	
	int numClients = 0;

	// default files directory
	//String filesDirName = "files";

    //File filesDir;
	
	ClientSettingsFile mSettings = new ClientSettingsFile();
	
	static String hostsFileName = "servers.dat";
	static String serverListTextFileName = "servers.txt";
	ServerListFile sListFile;
	
	Vector serverVector;	// vector of Hosts
	
	boolean connected = false;		// connected to server?
	
	Host currentServer = null;
	
	TwoWayMessageSocket messageSock = null;	// socket connection to server
	
	// client info
	
	Host localAddress;
	
	FileHost fileListMessage;	// message to send to server on connect accept
	
	WindowCloseListener closeListener;
	MenuActionListener menuListener;
	
	JPanel infoPanel;	// contains HardenedCriminal info

	JPanel infoNamePanel;
	JPanel infoAddressPanel;
	
	
	JLabel nameLabel;
	JLabel addressLabel;
	
	//JPanel buttonPanel;
	//JPanel rescanButtonPanel;
	//JPanel serversToTextButtonPanel;
	//JButton rescanButton;
	//JButton serversToTextButton;
	//ButtonActionListener buttonListen;
	
	StatePanel stPanel;
	SearchPanel srchPanel;
	ServerListPanel sListPanel;
	
	SearchResultsTable srTable;
	DownloadTable dlTable;
	UploadTable ulTable;
	
	JPanel topPanel;
	
	JPanel topTopPanel;
	JPanel topBottomPanel;
	
	JPanel tablesPanel;
	JPanel transferTablesPanel;
	
	JFrame transferFrame;	// separate frame for transfers

	JMenu fileMenu = new JMenu( "file" );
	JMenuItem selectFilesDirectoryMenuItem = new JMenuItem( "select files directory" );
	JMenuItem rescanFilesMenuItem = new JMenuItem( "rescan files" );
	JMenuItem saveTextServerListMenuItem = new JMenuItem( "save text server list" );
	JMenuItem quitMenuItem = new JMenuItem( "quit" );

	JMenuBar mainMenuBar = new JMenuBar();
	// menu to control which windows are showing
	JMenu windowMenu = new JMenu( "window" );
	JMenuItem transfersMenuItem = new JMenuItem( "transfers" );

	
	/**
	 * Constructs a client.
	 *
	 * @param prt port client will listen on
	 * @param maxCli maximum number of other clients that can connect
	 * to this client simultaneously
	 */
	public Client( int prt, int maxCli ) {
		
		super("konspire..client");
		
		GuiMessage.setParent( this );	// take over parenting for gui messages	
		
		port = prt;
		
		//filesDir = new File( filesDirName );
		
		// get local host address
		try {
			InetAddress iAddress = InetAddress.getLocalHost();
			localAddress = new Host( iAddress.getHostAddress(), port );
			}
		catch( UnknownHostException e ) {
			System.out.println( "Failed to retrieve local host IP address." );
			System.exit(1);
			}
		
		// setup local file host message
		fileListMessage = new FileHost( localAddress, mSettings.getFilesDirectory() );
		if( ! fileListMessage.fileDirectoryExists() ) {
			GuiMessage.show( "directory \"" + mSettings.getFilesDirectory().getName() + "\" does not exist." );
			}
		
		maxClients = maxCli;
		maxConnections =  2 * maxClients;
		
		setSize(500, 500);
		
		getContentPane().setLayout( new BoxLayout( getContentPane(), BoxLayout.Y_AXIS ) );
		
		infoPanel = new JPanel();
		infoPanel.setLayout( new BoxLayout( infoPanel, BoxLayout.X_AXIS ) );
		
		infoNamePanel = new JPanel( new FlowLayout( FlowLayout.LEFT ) );
		infoAddressPanel = new JPanel( new FlowLayout( FlowLayout.RIGHT ) );
		
		
		nameLabel = new JLabel( "HardenedCriminal software" );
		addressLabel = new JLabel( "www.hardenedcriminal.n3.net" );
		
		
		infoNamePanel.add( nameLabel );
		infoAddressPanel.add( addressLabel );
		
		infoPanel.add( infoNamePanel );
		infoPanel.add( infoAddressPanel );
		
		topPanel = new JPanel();
		topTopPanel = new JPanel();
		topBottomPanel = new JPanel();
		
		topPanel.setLayout( new BoxLayout( topPanel, BoxLayout.Y_AXIS ) );
		topTopPanel.setLayout( new BoxLayout( topTopPanel, BoxLayout.X_AXIS ) );
		topBottomPanel.setLayout( new BoxLayout( topBottomPanel, BoxLayout.X_AXIS ) );
		
		topPanel.add( infoPanel );
		topPanel.add( topTopPanel );
		topPanel.add( topBottomPanel );
		
		// need to init this first so it can be passed into srchPanel
		srTable = new SearchResultsTable( this );
		
		stPanel = new StatePanel( this );
		srchPanel = new SearchPanel( srTable, this );
		sListPanel = new ServerListPanel();
		
		topTopPanel.add( stPanel );
		topTopPanel.add( sListPanel );
		
		topBottomPanel.add( srchPanel );
		
		
		tablesPanel = new JPanel();
		
		tablesPanel.setLayout( new BoxLayout( tablesPanel, BoxLayout.Y_AXIS ) );
		
		
		transferTablesPanel = new JPanel();
		
		transferTablesPanel.setLayout( new BoxLayout( transferTablesPanel, BoxLayout.Y_AXIS ) );
		
		dlTable = new DownloadTable();
		ulTable = new UploadTable();
		
		tablesPanel.add( srTable );
		transferTablesPanel.add( dlTable );
		transferTablesPanel.add( ulTable );
		
		
		
		getContentPane().add( topPanel );
		getContentPane().add( tablesPanel );
		
		transferFrame = new JFrame( "konspire transfers" );
		transferFrame.setSize( 400, 400 );
		transferFrame.getContentPane().add( transferTablesPanel );
		
		
		
		setJMenuBar( mainMenuBar );
		
		selectFilesDirectoryMenuItem.setMnemonic( 'd' );
		rescanFilesMenuItem.setAccelerator( KeyStroke.getKeyStroke( 'R', Event.CTRL_MASK, false ) );
		rescanFilesMenuItem.setMnemonic( 'r' );
		selectFilesDirectoryMenuItem.setAccelerator( KeyStroke.getKeyStroke( 'D', Event.CTRL_MASK, false ) );
		saveTextServerListMenuItem.setMnemonic( 'l' );
		saveTextServerListMenuItem.setAccelerator( KeyStroke.getKeyStroke( 'L', Event.CTRL_MASK, false ) );
		quitMenuItem.setMnemonic( 'q' );
		quitMenuItem.setAccelerator( KeyStroke.getKeyStroke( 'Q', Event.CTRL_MASK, false ) );
		
		fileMenu.add( selectFilesDirectoryMenuItem );
		fileMenu.add( rescanFilesMenuItem );
		fileMenu.add( saveTextServerListMenuItem );
		fileMenu.addSeparator();
		fileMenu.add( quitMenuItem );
		
		fileMenu.setMnemonic( 'f' );
		
		
		transfersMenuItem.setMnemonic( 't' );
		transfersMenuItem.setAccelerator( KeyStroke.getKeyStroke( 'T', Event.CTRL_MASK, false ) );
		
		windowMenu.add( transfersMenuItem );
		windowMenu.setMnemonic( 'w' );
		
		mainMenuBar.add( fileMenu );
		mainMenuBar.add( windowMenu );
		
		
		menuListener = new MenuActionListener();
		transfersMenuItem.addActionListener( menuListener );
		selectFilesDirectoryMenuItem.addActionListener( menuListener );
		rescanFilesMenuItem.addActionListener( menuListener );
		saveTextServerListMenuItem.addActionListener( menuListener );
		quitMenuItem.addActionListener( menuListener );
		
		closeListener = new WindowCloseListener();
		
		addWindowListener( closeListener );
		
		
		// load server list from file
		sListFile = new ServerListFile( hostsFileName );
		serverVector = sListFile.readHostsVectorFromFile();
		sListPanel.setServerList( serverVector );
		
		show();
		transferFrame.show();
		}

    /**
     * Launches a dialog to select a directory.  Returns only after directory
     * is selected, or user cancels operation.
     *
     * @param inDirectory the directory to display by default when dialog opens
     * 
     * @return file selected, or null if user cancels operation
	 */
	private File launchDirectorySelector( File inDirectory ) {
		JFileChooser chooser;
		if( inDirectory != null ) {
			chooser = new JFileChooser( inDirectory );
			}
		else {
			chooser = new JFileChooser();	// else use default directory
			}
		chooser.setFileSelectionMode( chooser.DIRECTORIES_ONLY );
		
		int returnVal = chooser.showOpenDialog( mainClient );
		if( returnVal == JFileChooser.APPROVE_OPTION ) {
			return chooser.getSelectedFile();
			}
		else return null;
		
		}

    /**
     * Launches a dialog to select a location to save a file.  Returns only after file
     * is selected, or user cancels operation.
     *
     * @param inDirectory the directory to display by default when dialog opens
     * 
     * @return file selected, or null if user cancels operation
	 */
	private File launchSaveAs( File inDirectory ) {
		JFileChooser chooser;
		if( inDirectory != null ) {
			chooser = new JFileChooser( inDirectory );
			}
		else {
			chooser = new JFileChooser();	// else use default directory
			}
		chooser.setFileSelectionMode( chooser.FILES_ONLY );
	
		int returnVal = chooser.showSaveDialog( mainClient );
		if( returnVal == JFileChooser.APPROVE_OPTION ) {
	   		return chooser.getSelectedFile();
	   		}
	   	else return null;
		}
		
	
	/**
	 * Rescan files directory and resend file list.
	 */
	private void rescanFiles() {
		fileListMessage = new FileHost( localAddress, mSettings.getFilesDirectory() );
		if( ! fileListMessage.fileDirectoryExists() ) {
			GuiMessage.show( "directory \"" + mSettings.getFilesDirectory().getName() + "\" does not exist." );
			}
		// if connected, send new file list to server
		if( connected ) {
			try {
				messageSock.send( fileListMessage );
				}
			catch( IOException ioe ) {
				System.out.println( "Connection broken to server: " + ioe.toString() );
	            serverConnectionBroken();
				}
			}	
		}
	
	/**
	 * Listener for menu events.
	 */
	private class MenuActionListener implements ActionListener {
		public void actionPerformed( ActionEvent e ) {
			if( e.getSource() == transfersMenuItem ) {
				transferFrame.show();
				}
			else if( e.getSource() == selectFilesDirectoryMenuItem ) {
				//File newDir = launchDirectorySelector( mSettings.getFilesDirectory().getParentFile() );
				// changed for 1.1 compatability
				String parentFilePath = mSettings.getFilesDirectory().getParent();
				System.out.println( parentFilePath );
				File startDir = null;
				if( parentFilePath != null ) {
					startDir = new File( parentFilePath );
					} 
				File newDir = launchDirectorySelector( startDir );
				if( newDir != null ) {
					mSettings.setFilesDirectory( newDir );
					rescanFiles();
					}
				}
			else if( e.getSource() == rescanFilesMenuItem ) {
				rescanFiles();
				}
			else if( e.getSource() == saveTextServerListMenuItem ) {
				//File textServerListFile = launchSaveAs( mSettings.getFilesDirectory().getParentFile() );
				// changed for 1.1 compatability
				String parentFilePath = mSettings.getFilesDirectory().getParent();
				System.out.println( parentFilePath );
				File startDir = null;
				if( parentFilePath != null ) {
					startDir = new File( parentFilePath );
					} 
				File textServerListFile = launchSaveAs( startDir );
				if( textServerListFile != null ) {
					sListFile.writeHostsVectorToTextFile( serverVector, textServerListFile );
					}
				}
			else if( e.getSource() == quitMenuItem ) {
				System.exit(0);
				}
			}
		}
	
	
	/**
	 * Main client run loop.  Trys to connect to a server.
	 * Listens for incoming connections and creates
	 * a ReceiveThread for each one.
	 */
	public void run() {
		
		// first, try connecting to a server from the server list
		if( serverVector.size() >= 0 ) {
			// spawn a thread to try the connection
			ConnectThread cThread = new ConnectThread();
			cThread.start();
			}
		
		
		ServerSocket servSock;
		try {
			servSock = new ServerSocket( localAddress.getPort(), maxConnections );
			}
		catch( IOException e ) {
			System.out.println("Server socket failed to initialize.");
			return;
			}
		
		while( true ) {
			try {
				System.out.println("Client: Listening for connections." );
				Socket sock = servSock.accept();
				System.out.println( "Client: Connection received." );
				
				// connection received
				// pass it to receiver thread
				ReceiveThread rThread = new ReceiveThread( sock, this );
				
				rThread.start();
				
				// collect garbage
				System.gc();
				}
			catch( IOException e ) {
				System.out.println("Connection broken on server accept.");
				}
			}
		}	// end run()
	
	
    static Client mainClient;
	
	/**
	 * Constructs and launches a client at launch.
	 *
	 * @param args array of arguments passed in at launch
	 */
	public static void main( String args[] ) {
		
		randGen.setSeed( System.currentTimeMillis() );
		
		int cPort = defaultPort;
		int maxCli = defaultMaxClients;
		
		if( args.length >= 1 ) {
			// port passed in
			try{
				cPort = Integer.parseInt(args[0]);
				}
			catch( NumberFormatException e ) {
				usage();
				}
			}
		if( args.length >= 2 ) {
			// max clients passed in
			try{
				maxCli = Integer.parseInt(args[1]);
				}
			catch( NumberFormatException e ) {
				usage();
				}
			}
		
			
		mainClient = new Client( cPort, maxCli );
		
		mainClient.run();

		}
	/**
	 * Prints usage help message and exits.
	 */
	private static void usage() {				

		System.out.print("usage:\n");
		System.out.print("Client [port_number] [max_connect]\n");
		
		System.out.print("examples:\n\t");
		System.out.print("Client\n\t");
		System.out.print("Client 6433\n\t");
		System.out.print("Client 6433 15\n\t");
		
		System.exit(1);
		}
	
	// implements the download controller interface
	
	public void startDownload( FileDescriptor f ) {
		DownloadThread dlThread = new DownloadThread( f, mSettings.getFilesDirectory() );
		dlTable.addThread( dlThread );
		}
	
	/**
	 * Called by components if connection to server breaks.
	 */
	public void serverConnectionBroken() {
		
		serverVector.removeElement( currentServer );
		currentServer = null;
		messageSock = null;
		connected = false;
		
		srchPanel.setSearchSocket( messageSock );
		stPanel.setLiveServer( currentServer );
		
		ConnectThread cThread = new ConnectThread();
		cThread.start();
		}
	
	// should server vector be obtained from next server connected to?
	boolean serverVectorNeedsUpdate = true;
	boolean serverVectorBeenUpdated = false;
	// this lets loop through servers know whether to restart enumeration or not
	
	
	/**
	 * Tries to connect to a server on the server list.
	 * Tries all until list exhausted.
	 *
	 * @return true iff connected to a server on the list successfully
	 */
	public boolean tryConnecting() {
		shuffleVector( serverVector );
		Enumeration em = serverVector.elements();
		try {
	
			Host failedHost = localAddress;	// set to dummy address for now
			
			serverVectorNeedsUpdate = true;
			serverVectorBeenUpdated = false;
			
			while( em.hasMoreElements() ) {
				
				Host s = (Host)( em.nextElement() );
				
				if( connectToServer( s ) ) {
					currentServer = s;
					connected();
					return true;
					}
				else {	// need to refresh enumeration due to new server vector
					if( serverVectorNeedsUpdate && serverVectorBeenUpdated ) {
						em = serverVector.elements();
						serverVectorNeedsUpdate = false;
						}
					}					
				}
			}
		catch( NoSuchElementException ex ) {
			System.out.println( "Iterating through server vector failed." );
			}
		return false;	// no connection established
		}

	/**
	 * Tries to connect to a server.
	 *
	 * @return true iff connected to server successfully
	 */
	public boolean connectToServer( Host s ) {
		
		try {
			// create a new message socket
			messageSock = new TwoWayMessageSocket( s );
			
			// send connect request
			messageSock.send( new ClientRequestConnect( localAddress ) );
			
			// now receive server list from server
			Message m = messageSock.receive();
			
			if( m instanceof ServerList ) {
				ServerList sList = (ServerList) m;
				Host correctLocalAddress = sList.getReceiverCorrectAddress();
				if( !correctLocalAddress.equals( localAddress ) ) {
					// we retrieved or local address incorrectly
					localAddress = correctLocalAddress;
					fileListMessage.setHost( localAddress );
					System.out.println( "Received a corrected local address from server: " + localAddress );
					}
				if( serverVectorNeedsUpdate ) {
					serverVector = sList.getHostVector();
					sListPanel.setServerList( sList );
					serverVectorBeenUpdated = true;
					
					// write new server list out to file
					sListFile.writeHostsVectorToFile( serverVector );
					}
				processStats( sList );
				}
			else {
				unexpectedServerMessage( m );
				messageSock.close();
				messageSock = null;
				return false;			
				}
			
			// next receive connection authorization from server
			m = messageSock.receive();
			if( m instanceof ConnectResponse ) {
				ConnectResponse cr = (ConnectResponse) m;
				if( cr.isAccepted() ) {
					// accepted by this server
					
					// send our file list
					messageSock.send( fileListMessage );
					
					return true;
					}
				else {
					// rejected
					messageSock.close();
					messageSock = null;
					return false;
					}
				}
			else {
				unexpectedServerMessage( m );
				messageSock.close();
				messageSock = null;
				return false;			
				}
			}
		catch( IOException e ) {
			System.out.println( "Connection broken to server: " + e.toString() );
			messageSock = null;
			return false;
			}
		
		
		}
	
	/**
	 * Informs various components about current connection.
	 */
	private void connected() {
		srchPanel.setSearchSocket( messageSock );
		stPanel.setLiveServer( currentServer );
		connected = true;
		}
	
	/**
	 * Handles receipt of an unexpected message type.
	 *
	 * @param m message that was of an unexpected type
	 */	
	private void unexpectedServerMessage( Message m ) {
		System.out.println( "Server responded to search with an unexpected message type: " + m.getClass().toString() );
		}
	
	/**
	 * Handles receipt of a message that contains updated system stats.
	 *
	 * @param m message sent to client containing updated system stats
	 */	
	public void processStats( ToClientMessage m ) {
		stPanel.setOnlineStats( m );
		}
	
	/**
	 * Tries to connect to seed server.
	 *
	 * @param s <code>Host</code> to connect to as a seed server
	 */
	public void trySeed( Host s ) {
		
		serverVector.addElement( s );
		
		if( !connected ) {
			ConnectThread cThread = new ConnectThread();
			cThread.start();
			}
		}
	
	/**
	 * Called by client handling thread whenever a client disconnects.
	 */
	public void clientDisconnected() {
		numClients--;
		}
	
	
	/**
	 * Shuffles a <code>Vector</code> by switching its starting point.
	 *
	 * @param v <code>Vector</code> to shuffle
	 */
	private void shuffleVector( Vector v ) {
		Object array[] = v.toArray();
		
		if( array.length > 0 ) {
			v.removeAllElements();
			
			int startIndex = (int)( Math.round( randGen.nextFloat() * ( array.length - 1 ) ) );
			
			for( int i=startIndex; i< array.length; i++ ) {
				v.addElement( array[i] );	
				}
			
			for( int i=0; i<startIndex; i++ ) {
				v.addElement( array[i] );
				}
			}	
		}
	

	/**
	 * Listens for window close event
	 */
	private class WindowCloseListener extends DefaultWindowListener {
	
		public void windowClosing( WindowEvent e ) {
			System.exit(0);
			}	
		}

	
	/**
	 * Thread that handles establishing a connection to a server from the server list
	 */
	class ConnectThread extends Thread {
		
		/**
		 * Constructs a <code>ConnectThread</code>.
		 */	
		public ConnectThread() {
			setPriority( Thread.MIN_PRIORITY );
			}
		
		public void run() {
			boolean connectionEstablished = tryConnecting();
			if( !connectionEstablished ) {
				stPanel.setLiveServer( null );
				// empty out server list, since none are valid
				// this is done so user not confused
				// user must enter a new seed server to connect
				serverVector = new Vector();
				sListPanel.setServerList( serverVector );
				GuiMessage.show( "connections to all servers on server list failed." );
				}
			}
		}
		
			
	/**
	 * Thread for receiving and processing initial messages from a 
	 * newly received socket connection.
	 * <p>
     * Spawns other threads as appropriate.
     */
	class ReceiveThread extends Thread {
	
		Socket sock;

		TwoWayMessageSocket messageSock;
	
		Client parent;
		
		/**
		 * Constructs a <code>ReceiveThread</code>
		 *
		 * @param s the socket that connection has arrived on
		 * @param par the client parent of this thread
		 */
		public ReceiveThread( Socket s, Client par ) {
			sock = s;
			parent = par;
			setPriority( Thread.MIN_PRIORITY );
			}
	
		public void run() {
			
			try {
				messageSock = new TwoWayMessageSocket( sock );
				}
			catch( IOException ioe ) {
				System.out.println("Exception: " + ioe.toString() );
				System.out.println("Connection broken while receive thread setting up messsage socket.");
				return;
				}
			
			try {
				Message m = messageSock.receive();
				
				if( m instanceof ClientFileRequest ) {
					// client wants to dl a file
					ClientFileRequest fileReq = (ClientFileRequest) m;
					
					boolean accept = false;
					
					if( numClients < maxClients ) {
						// accept client connection
						numClients++;
						accept = true;
						}
					
					// create thread and add it to the upload table	
					UploadThread ulThread = new UploadThread( fileReq, mSettings.getFilesDirectory(), accept, messageSock, parent );
					ulTable.addThread( ulThread );
					return;
					}
				// ignore other message types
				System.out.println( "Client received an unexpected message type: " + m.getClass().toString() );
				messageSock.closeTotal();
				return;
				}
			catch( IOException ioe ) {
				System.out.println("Exception: " + ioe.toString() );
				System.out.println("Connection broken while receive thread receiving message.");
				return;
				}
			}	// end run()
		}	// end inner class ReceiveThread		
				
	}	// end class client

		