// Jason Rohrer
// ClientSettingsFile.java

/**
*
*	Object for sending client settings to file.
*
*	Created 7-4-2000
*	Mods:
*/


package konspire.client;

import java.io.*;


/**
 * Object for sending client settings to file.  Contains default settings for client.
 *
 * @author Jason Rohrer
 */
public class ClientSettingsFile {

	
	String mFileName = "ClientSettings.dat";
	
	ClientSettings mClientSettings = new ClientSettings();

	
	/**
	 * Construct a new ClientSettingsFile and read settings from file
	 *
	 */
	public ClientSettingsFile() {
		readFromFile();
		}
	
	
	/**
	 * Set files directory.
	 *
	 * @param inFilesDirector the new files directory
	 *
	 */
	public void setFilesDirectory( File inFileDirectory ) {
	
		mClientSettings.mFilesDirectory = inFileDirectory;
		
		writeToFile();
		}
	
	/**
	 * Get files directory.
	 *
	 * @return the files directory
	 *
	 */
	public File getFilesDirectory() {
	
		return mClientSettings.mFilesDirectory;
		
		}
		
	
	/**
	 * Write settings to file.
	 *
	 */	
	private void writeToFile() {
		try {
			DataOutputStream fileOutput = new DataOutputStream (
							new FileOutputStream( mFileName ) );
			
			ObjectOutputStream objOutStream 
					= new ObjectOutputStream( fileOutput );
			
			// write the settings object to file
			objOutStream.writeObject( mClientSettings );
			
			objOutStream.flush();
			objOutStream.close();
			fileOutput.close();
			}
		catch( IOException e ) {
			System.out.println( "Writing settings to file " + mFileName + " failed," + e );
			}
		}
		
	/**
	 * Read settings from file.
	 *
	 */	
	private void readFromFile() {
		File testExistsFile = new File( mFileName );
		if( ! testExistsFile.exists() ) {
			writeToFile();
			}
		
		try {
			DataInputStream fileInput = new DataInputStream (
							new FileInputStream( mFileName ) );
			
			ObjectInputStream objInStream 
					= new ObjectInputStream( fileInput );
			
			try {
				Object inObject = objInStream.readObject();

				if( inObject instanceof ClientSettings ) {
					mClientSettings = (ClientSettings)inObject;
					}
				objInStream.close();			
				fileInput.close();		
				}
			catch( ClassNotFoundException ex ) {
				System.out.println( "Unknown object read from settings file: " + mFileName );
				fileInput.close();
				}	
			}
		catch( IOException e ) {
			System.out.println( "Reading settings from file " + mFileName + " failed." );
			}
		}		 
	}


