// Jason Rohrer
// FileDescriptorGUI.java

/**
*
*	GUI wrapper for a file descriptor (attaches to a spot in a scroll table
*	for listing search results)
*
*	Created 5-10-2000
*	Mods:
*		Jason Rohrer	7-3-2000	Changed to use UnitsPrefixFormat to format file size.
*/

package konspire.client;

import konspire.common.FileDescriptor;

import konspire.common.UnitsPrefixFormat;
import konspire.common.BaseTwoUnitsPrefixFormat;
import java.text.NumberFormat;


import javax.swing.JButton;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;


/**
 * GUI wrapper for a file descriptor. Attaches to a spot in a 
 * {@link SearchResultsTable}.
 *
 * @author Jason Rohrer
 */
public class FileDescriptorGUI {

	FileDescriptor f;
	
	double fileSizeInUnits = 0;
	String baseUnitsString = "B";
	
	
	JButton dlButton;
	
	SearchResultsTable sRTable;
	
	TableRow row;	// pointer to row in table where this component is represented
	
	ButtonActionListener buttonListen;
	
	
	/**
	 * Constructs a GUI from a <cod>FileDescriptor</code> and a <code>SearchResultsTable</code>
	 *
	 * @param fDes the file descriptor this GUI will represent
	 * @param table the table where this GUI will display itself
	 */
	public FileDescriptorGUI( FileDescriptor fDes, SearchResultsTable table ) {
		
		f = fDes;
		
		UnitsPrefixFormat prefixFormat = new BaseTwoUnitsPrefixFormat();
		String fileSizeString = prefixFormat.convertUnitsPrefix( f.getSize(), 
									baseUnitsString );
		
		
		sRTable = table;
		
		dlButton = new JButton( "download" );
		
		Object [] oArray = { f.getName(), fileSizeString, dlButton };
		
		row = sRTable.addRow( oArray );  
		
		buttonListen = new ButtonActionListener();
		
		dlButton.addActionListener( buttonListen );
		}  
	
	
	/**
	 * Listener for button actions in this GUI.
	 *
	 */
	private class ButtonActionListener implements ActionListener {
		public void actionPerformed( ActionEvent e ) {
			if( e.getSource() == dlButton ) {
				
				// user chose to download this file
				
				sRTable.removeRow( row );
				
				sRTable.startDownload( f );
				}
			}
		}
	
	
	}